import logging

import pandapower as pp

from ..constraints.load_voltage_change import ConstraintLoadVoltageChange
from .base import GridElement

LOG = logging.getLogger(__name__)


class PPLoad(GridElement):
    @staticmethod
    def pp_key() -> str:
        return "load"

    @staticmethod
    def res_pp_key() -> str:
        return "res_load"

    def __init__(self, index, grid, value=0.02):
        super().__init__(index, grid, LOG)

        self.current_bus_voltage = 1.0
        self.in_service = True
        self.p_mw = grid.get_value(self.pp_key(), index, "p_mw")
        self.q_mvar = grid.get_value(self.pp_key(), index, "q_mvar")

        self._constraints.append(ConstraintLoadVoltageChange(self, value))

    def step(self, time):
        old_state = self.in_service
        self.in_service = True
        self._check(time)

        self.set_value("p_mw", self.p_mw)
        self.set_value("q_mvar", self.q_mvar)
        self.set_value("in_service", self.in_service)

        # Run powerflow so that the next unit does not automatically
        # switch off, too.
        if old_state != self.in_service:
            if not self.in_service:
                LOG.debug(
                    f"At step {time}: Load {self.index} with p={self.p_mw:.5f}"
                    f" and q={self.q_mvar:.5f} out of service (Bus voltage: "
                    f"{self.current_bus_voltage:.5f})"
                )
                try:
                    self.grid.run_powerflow()
                except pp.LoadflowNotConverged:
                    LOG.debug("Line disabled. PF still not converging.")
            else:
                LOG.debug(
                    f"At step {time}: Load {self.index} back in service."
                )
                try:
                    self.grid.run_powerflow()
                except pp.LoadflowNotConverged:
                    LOG.debug("Load re-enabled. PF not converging.")

        return old_state != self.in_service



import logging

import pandapower as pp

from ..constraints.line_loading import ConstraintLineLoading
from .base import GridElement

LOG = logging.getLogger(__name__)


class PPLine(GridElement):
    @staticmethod
    def pp_key() -> str:
        return "line"

    @staticmethod
    def res_pp_key() -> str:
        return "res_load"

    def __init__(self, index, grid, value=100):
        super().__init__(index, grid, LOG)

        self.in_service = True
        self.max_percentage = value

        self._constraints.append(ConstraintLineLoading(self))

    def step(self, time):
        old_state = self.in_service
        self.in_service = True
        self._check(time)
        self.set_value("in_service", self.in_service)
        if not old_state and self.in_service:
            self.grid.run_powerflow()
            self._check(time)
            self.set_value("in_service", self.in_service)
            self.grid.run_powerflow()

        if old_state != self.in_service:
            if not self.in_service:
                LOG.debug(f"At step {time}: Line {self.index} out of service.")
                try:
                    self.grid.run_powerflow()
                except pp.LoadflowNotConverged:
                    LOG.debug("Line disabled. PF still not converging.")
            else:
                LOG.debug(
                    f"At step {time}: Line {self.index} back in service."
                )
                try:
                    self.grid.run_powerflow()
                except pp.LoadflowNotConverged:
                    LOG.debug("Line re-enabled. PF not converging.")

        return old_state != self.in_service



import logging

from .base import GridElement

LOG = logging.getLogger(__name__)


class PPSwitch(GridElement):
    @staticmethod
    def pp_key() -> str:
        return "switch"

    @staticmethod
    def res_pp_key() -> str:
        return "res_switch"

    def __init__(self, index, grid, value):
        super().__init__(index, grid, LOG)
        self.closed = True

    def step(self, time):
        old_state = self.closed
        self.closed = True
        self._check(time)

        self.set_value("closed", self.closed)

        return old_state != self.closed



import os
from os import path

import matplotlib.pyplot as plt
from pandapower import plotting

CIGRE_CFG = {
    "slack_size": 0.25,
    "slack_color": "black",
    "slack_zorder": 3,
    "trafo_plot": True,
    "trafo_size": 0.2,
    "trafo_color": "black",
    "trafo_zorder": 0,
    "bus_size": 0.2,
    "bus_color": "blue",
    "bus_zorder": 2,
    "line_size": 2,
    "line_color": "black",
    "line_zorder": 1,
}

SIMBENCH_CFG = {
    "slack_size": 0.00025,
    "slack_color": "black",
    "slack_zorder": 3,
    "trafo_plot": False,
    "trafo_size": 0.0002,
    "trafo_color": "black",
    "trafo_zorder": 0,
    "bus_size": 0.0002,
    "bus_color": "blue",
    "bus_zorder": 2,
    "line_size": 1,
    "line_color": "black",
    "line_zorder": 1,
}


class Plotter:
    """Midas grid plotting class"""

    def __init__(self):
        self.grid = None
        self.grid_type = None
        self.plot_path = "."
        self.plot_name = "DefaultName"
        self.cfg = CIGRE_CFG

    def simple_plot(self):
        """Plot the grid without heatmap"""
        plotting.simple_plot(self.grid)

    def _get_plot(self, figsize):
        cmaps, norms = self._create_cmaps(self.cfg)

        gridplot = plotting.draw_collections(
            self._create_collections(self.cfg, cmaps, norms), figsize
        )

        return gridplot

    def plot(self, name, current_idx=0, figsize=(16, 9)):
        """Plot the grid with heatmaps"""

        if self.grid_type is None or self.grid_type == "cigre":
            self.cfg = CIGRE_CFG
        elif self.grid_type == "simbench":
            self.cfg = SIMBENCH_CFG

        self._get_plot(figsize)
        os.makedirs(self.plot_path, exist_ok=True)
        output = path.join(
            self.plot_path, f"{self.plot_name}-{name}-{current_idx:05d}.png"
        )

        plt.savefig(output)
        plt.close()

    def _create_cmaps(self, cfg):
        """Create pandapower cmaps for plotting."""

        if cfg["trafo_plot"]:
            # Create color map for trafo
            colors_trafo = [(0, "green"), (50, "yellow"), (100, "red")]
            cmap_trafos, norm_trafos = plotting.cmap_continuous(colors_trafo)
        else:
            cmap_trafos = None
            norm_trafos = None

        # Create color map for node voltage
        colors_buses = [(0.935, "blue"), (1.0, "green"), (1.065, "red")]
        cmap_buses, norm_buses = plotting.cmap_continuous(colors_buses)

        # Create color map for line loadings
        colors_lines = [(0, "green"), (40, "yellow"), (80, "red")]
        cmap_lines, norm_lines = plotting.cmap_continuous(colors_lines)

        return (cmap_trafos, cmap_buses, cmap_lines), (
            norm_trafos,
            norm_buses,
            norm_lines,
        )

    def _create_collections(
        self, cfg, cmaps=(None, None, None), norms=(None, None, None)
    ):
        """Create pandapower collections."""

        slack = plotting.create_bus_collection(
            self.grid,
            self.grid.ext_grid.bus.values,
            size=cfg["slack_size"],
            color=cfg["slack_color"],
            zorder=cfg["slack_zorder"],
            patch_type="rect",
        )

        trafo_sec, trafo_prim = self._create_trafo_collection(
            cfg, cmaps[0], norms[0]
        )

        buses = self._create_bus_collection(cfg, cmaps[1], norms[1])

        lines, open_lines = self._create_line_collection(
            cfg, cmaps[2], norms[2]
        )

        cols = [slack, trafo_sec, trafo_prim, buses, lines]

        if open_lines is not None:
            cols.append(open_lines)
        return cols

    def _create_trafo_collection(self, cfg, cmap, norm):
        """Return pandapower trafo collection."""
        if cmap is not None and norm is not None:
            trafo_sec, trafo_prim = plotting.create_trafo_collection(
                self.grid,
                self.grid.trafo.index,
                cmap=cmap,
                norm=norm,
                size=cfg["trafo_size"],
                zorder=cfg["trafo_zorder"],
            )
        else:
            trafo_sec, trafo_prim = plotting.create_trafo_collection(
                self.grid,
                self.grid.trafo.index,
                color=cfg["trafo_color"],
                size=cfg["trafo_size"],
                zorder=cfg["trafo_zorder"],
            )

        return trafo_sec, trafo_prim

    def _create_bus_collection(self, cfg, cmap, norm):
        """Return pandapower bus collection."""
        if cmap is not None and norm is not None:
            buses = plotting.create_bus_collection(
                self.grid,
                self.grid.bus.index,
                size=cfg["bus_size"],
                cmap=cmap,
                norm=norm,
                z=self.grid.res_bus.vm_pu,
                zorder=cfg["bus_zorder"],
            )
        else:
            buses = plotting.create_bus_collection(
                self.grid,
                self.grid.bus.index,
                size=cfg["bus_size"],
                zorder=cfg["bus_zorder"],
            )

        return buses

    def _create_line_collection(self, cfg, cmap, norm):
        """Return pandapower line collection."""

        open_switch_index = self.grid.switch.loc[
            (self.grid.switch["et"] == "1") & ~(self.grid.switch["closed"])
        ]["element"]

        closed_ids = self.grid.line.loc[
            ~self.grid.line.index.isin(open_switch_index)
        ]
        open_ids = self.grid.line.loc[
            self.grid.line.index.isin(open_switch_index)
        ]

        if cmap is not None and norm is not None:
            lines = plotting.create_line_collection(
                self.grid,
                closed_ids.index,
                use_bus_geodata=True,
                cmap=cmap,
                norm=norm,
                linewidths=cfg["line_size"],
                zorder=cfg["line_zorder"],
            )
        else:
            lines = plotting.create_line_collection(
                self.grid,
                closed_ids.index,
                use_bus_geodata=True,
                color=cfg["line_color"],
                linewidths=cfg["line_size"],
                zorder=cfg["line_zorder"],
            )

        open_lines = plotting.create_line_collection(
            self.grid,
            open_ids.index,
            use_bus_geodata=True,
            line_style=":",
            color=cfg["line_color"],
            linewidths=cfg["line_size"],
            zorder=cfg["line_zorder"],
        )

        return lines, open_lines



import numpy as np

from .base import Constraint


class ConstraintLineLoading(Constraint):
    def __init__(self, element):
        super().__init__(element)

    def check(self, time) -> bool:
        self._satisfied = True
        self.expected_value = self._element.max_percentage

        self.violated_value = self._element.grid.get_value(
            "res_line", self._element.index, "loading_percent"
        )
        if np.isnan(self.violated_value):
            # NaN occurs when LF fails or the line is not connected
            # It is assumed that there is no violation then
            self.violated_value = 0
        if self.violated_value > self.expected_value:
            self._satisfied = False

        return self._satisfied

    def handle_violation(self):
        self._element.in_service = False



"""This module contains the test cases for the pandapower simulator."""

import os
import unittest

from midas_powergrid.simulator import PandapowerSimulator


class TestSimulator(unittest.TestCase):
    """Test cases for the pandapower mosaik simulator."""

    def test_init(self):
        """Test the simulator initialization."""
        sim = PandapowerSimulator()
        meta = sim.init(
            sid="TestSimulator-0", step_size=900, sim_params=dict()
        )

        self.assertTrue("api_version" in meta)
        self.assertTrue("models" in meta)
        self.assertTrue("Grid" in meta["models"])

    def test_create(self):
        """Test the creation of a model."""

        sim = PandapowerSimulator()
        sim.init(sid="TestSimulator-0", step_size=900, sim_params=dict())

        entities = sim.create(
            num=2, model="Grid", gridfile="simple_four_bus_system"
        )

        self.assertEqual(len(entities), 2)

        for idx, entity in enumerate(entities):
            self.assertEqual(len(entity), 4)
            self.assertEqual(entity["eid"], f"Grid-{idx}")
            self.assertEqual(len(entity["children"]), 11)

    def test_step(self):
        """Test the step of the simulator.

        Two grids are instantiated at the same time to see that the
        inputs are set to the correct grid.
        """
        sim = PandapowerSimulator()
        sim.init(sid="TestSimulator-0", step_size=900, sim_params=dict())
        sim.create(2, "Grid", gridfile="simple_four_bus_system")

        inputs = {
            "0-load-0-2": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-0": 0.02,
                    "DummySim-0.DummyHousehold-1": 0.02,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-0": 0.01,
                    "DummySim-0.DummyHousehold-1": 0.005,
                },
            },
            "1-load-1-3": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-2": 0.03,
                    "DummySim-0.DummyHousehold-3": 0.01,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-2": 0.005,
                    "DummySim-0.DummyHousehold-3": 0.01,
                },
            },
        }

        sim.step(0, dict())
        self.assertAlmostEqual(
            sim.models["Grid-0"].grid.get_value("res_bus", 1, "vm_pu"),
            0.996608,
        )
        self.assertAlmostEqual(
            sim.models["Grid-1"].grid.get_value("res_bus", 1, "vm_pu"),
            0.996608,
        )

        sim.step(0, inputs)
        self.assertAlmostEqual(
            sim.models["Grid-0"].grid.get_value("res_bus", 1, "vm_pu"),
            0.9952088,
        )
        self.assertAlmostEqual(
            sim.models["Grid-1"].grid.get_value("res_bus", 1, "vm_pu"),
            0.9950847,
        )

    def test_get_data(self):
        """Test the get data method of the simulator."""
        sim = PandapowerSimulator()
        sim.init(sid="TestSimulator-0", step_size=900, sim_params=dict())
        sim.create(2, "Grid", gridfile="simple_four_bus_system")
        sim.step(0, dict())

        outputs = {
            "0-bus-1": ["vm_pu", "va_degree"],
            "1-bus-2": ["vm_pu", "va_degree"],
            "0-line-0": ["loading_percent"],
            "Grid-0": ["health", "grid_json"],
        }

        data = sim.get_data(outputs)

        self.assertAlmostEqual(data["0-bus-1"]["vm_pu"], 0.996608)
        self.assertAlmostEqual(data["1-bus-2"]["vm_pu"], 0.9377604)
        self.assertAlmostEqual(
            data["0-line-0"]["loading_percent"], 31.273, places=3
        )
        self.assertAlmostEqual(data["Grid-0"]["health"], 0.9454563)

    def test_with_included_slack_bus(self):
        sim = PandapowerSimulator()
        sim.init(sid="TestSimulator-0", step_size=900, sim_params={})
        sim.create(1, "Grid", gridfile="midasmv", include_slack_bus=True)
        sim.step(0, {})

        outputs = {
            "0-bus-0": ["vm_pu", "va_degree"],
            "0-bus-1": ["vm_pu", "va_degree"],
        }

        data = sim.get_data(outputs)

        self.assertAlmostEqual(data["0-bus-0"]["vm_pu"], 1.03)
        self.assertAlmostEqual(data["0-bus-1"]["vm_pu"], 1.0314206)

        self.assertAlmostEqual(data["0-bus-0"]["va_degree"], 0.0)
        self.assertAlmostEqual(data["0-bus-1"]["va_degree"], -30.00117963)

    def test_with_plotting(self):
        """Test if plotting works as expected."""
        plot_file = os.path.abspath(os.path.join(__file__, "..", "_plots"))
        sim = PandapowerSimulator()
        sim.init(
            sid="TestSimulator-0",
            step_size=900,
            plotting=True,
            plot_path=plot_file,
        )
        sim.create(1, "Grid", gridfile="simple_four_bus_system", plotting=True)

        sim.step(0, dict())

        self.assertTrue(os.path.isdir(plot_file))
        files = os.listdir(plot_file)
        self.assertEqual(1, len(files))
        self.assertTrue(files[0].endswith(".png"))

        # Clean up
        os.remove(os.path.join(plot_file, files[0]))
        os.rmdir(plot_file)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from ..elements.base import GridElement
from .base import Constraint


class ConstraintVoltageBand(Constraint):
    def __init__(
        self, element: GridElement, lower_band: float, upper_band: float
    ):
        super().__init__(element)

        self._vm_pu_max = upper_band
        self._vm_pu_min = lower_band

        self.over_voltage = False
        self.under_voltage = False

    def check(self, time) -> bool:
        self._satisfied = True
        voltage = self._element.grid.get_value(
            "res_bus", self._element.index, "vm_pu"
        )

        self.over_voltage = voltage > self._vm_pu_max
        self.under_voltage = voltage < self._vm_pu_min

        if np.isnan(voltage) or self.over_voltage or self.under_voltage:
            self._satisfied = False
            self.violated_value = voltage

        return self._satisfied

    def handle_violation(self):
        self._element.set_load_service_state(in_service=False)
        self._element.set_sgen_service_state(in_service=False)

        self._element.in_service = False



import csv
import os
from os.path import join

import joblib
import numpy as np

from .pp_grid import PPGrid


class SurrogateGrid(PPGrid):
    def __init__(self, grid_name, grid_params):
        surrogate_params = grid_params["surrogate_params"]
        super().__init__(grid_name, grid_params)

        base_path = surrogate_params["base_path"]
        self._model = joblib.load(
            join(base_path, surrogate_params["model_path"])
        )
        self._x_scaler = joblib.load(
            join(base_path, surrogate_params["x_scaler"])
        )
        self._y_scaler = joblib.load(
            join(base_path, surrogate_params["y_scaler"])
        )
        self._input_map = joblib.load(
            join(base_path, surrogate_params["input_map"])
        )
        self._output_map = joblib.load(
            join(base_path, surrogate_params["output_map"])
        )
        self._evaluate = surrogate_params.get("evaluate", False)
        # self._evaluate = (
        #     True
        #     if surrogate_params.get("evaluate", "true") == "true"
        #     else False
        # )
        self._save_path = surrogate_params.get("save_path", ".")
        self._model_result_path = join(
            self._save_path,
            surrogate_params.get("save_model_results_to", "model_results.csv"),
        )
        self._ppg_results_path = join(
            self._save_path,
            surrogate_params.get("save_ppgrid_results_to", "ppg_results.csv"),
        )

        self._inputs = np.zeros((len(self._input_map),))
        self._outputs = np.zeros((len(self._output_map),))
        self._rmses = {
            "vm_pu": [],
            "va_degree": [],
            "bus_p_mw": [],
            "bus_q_mvar": [],
            "line_loading": [],
            "trafo_loading": [],
            "eg_p_mw": [],
            "eg_q_mvar": [],
        }
        self._model_results = []
        self._ppg_results = []

    def set_value(self, etype, idx, attr, val):
        key = f"{etype.capitalize()}_{idx:03d}__{attr}"
        self._inputs[self._input_map[key]] = val

        if self._evaluate:
            super().set_value(etype, idx, attr, val)

    def run_powerflow(self):
        self._outputs = self._y_scaler.inverse_transform(
            self._model.predict(
                self._x_scaler.transform(self._inputs.reshape(1, -1))
            )
        )[0]
        if self._inputs.sum() == 0.0 or self._evaluate:
            super().run_powerflow()

        if self._evaluate:
            self._model_results.append(list(self._outputs))
            self._ppg_results.append(
                list(self._grid.res_bus.vm_pu.values[1:])
                + list(self._grid.res_bus.va_degree.values[1:])
                + list(self._grid.res_bus.p_mw.values[1:])
                + list(self._grid.res_bus.q_mvar.values[1:])
                + list(self._grid.res_line.loading_percent.values)
                + list(self._grid.res_trafo.loading_percent.values)
                + list(self._grid.res_ext_grid.p_mw.values)
                + list(self._grid.res_ext_grid.q_mvar.values)
            )

    def get_value(self, etype, idx=None, attr=None):
        if idx is not None and attr is not None:
            key = f"{etype.capitalize()}_{idx:03d}__{attr}"
            if key in self._input_map:
                return self._inputs[self._input_map[key]]
            elif key in self._output_map:
                return self._outputs[self._output_map[key]]
            else:
                return super().get_value(etype, idx, attr)
        elif attr is not None:
            keys = [
                k
                for k in self._input_map
                if (etype.capitalize() in k and attr in k)
            ]
            if keys:
                return [self._inputs[self._input_map[k]] for k in keys]
            keys = [
                k
                for k in self._output_map
                if (etype.capitalize() in k and attr in k)
            ]
            if keys:
                return [self._output[self._output_map[k]] for k in keys]

        return super().get_value(etype, idx, attr)

    def finalize(self):
        if self._evaluate:
            with open(
                join(os.getcwd(), self._model_result_path), "w", newline=None
            ) as csv_file:
                writer = csv.writer(
                    csv_file,
                    delimiter=",",
                    quotechar='"',
                    quoting=csv.QUOTE_MINIMAL,
                )

                writer.writerow(list(self._output_map.keys()))
                for idx in range(len(self._model_results)):
                    row = self._model_results[idx]
                    writer.writerow(row)

            with open(
                join(os.getcwd(), self._ppg_results_path), "w", newline=None
            ) as csv_file:
                writer = csv.writer(
                    csv_file,
                    delimiter=",",
                    quotechar='"',
                    quoting=csv.QUOTE_MINIMAL,
                )

                writer.writerow(list(self._output_map.keys()))
                for idx in range(len(self._ppg_results)):
                    row = self._ppg_results[idx]
                    writer.writerow(row)

            # print(self._rmses)
            # print(
            #     f"VM PU: {np.average(self._rmses['vm_pu'])} VA Degree: "
            #     f"{np.average(self._rmses['va_degree'])} "
            # )


def rmse(v_true, v_pred):
    return np.sqrt((v_true - v_pred) ** 2)



from abc import abstractmethod

# from midas.core.powergrid.c import constraints


class GridElement:
    @staticmethod
    @abstractmethod
    def pp_key() -> str:
        """
        Returns
        -------
        str
            Returns pandapower's identifier for self element type
        """
        pass

    @staticmethod
    @abstractmethod
    def res_pp_key() -> str:
        """
        Returns
        -------
        str
            Returns pandapower's identifier for self res_element type
        """
        pass

    def __init__(self, index, grid_model, log):
        self.grid = grid_model
        self.index = index
        self.log = log

        self._constraints = list()
        self._unsatisfied_constraints = list()
        self._manipulations = list()

    def add_constraint(self, constraint):
        self._constraints.append(constraint)

    def add_constraints(self, constraints):
        self._constraints.extend(constraints)

    def remove_constraint(self, constraint):
        self._constraints.remove(constraint)

    def parameters(self):
        # columns = list(self.grid.)
        pass

    def set_value(self, attribute, value):
        self.grid.set_value(self.pp_key(), self.index, attribute, value)

    def step(self, time) -> bool:
        return False

    def _check(self, time):
        if self._constraints:
            self._unsatisfied_constraints = [
                constr
                for constr in self._constraints
                if constr.check(time) is False
            ]

            for constraint in self._unsatisfied_constraints:
                constraint.handle_violation()

            # if not self.parameters()["in_service"]:
            #     if self._can_reconnect():
            #         self.set_value("in_service", True)

    # def _can_reconnect(self) -> bool:
    # from grid.elements.PPBus import PPBus

    # bus = self.grid_model.grid_state[PPBus][self.bus]
    # voltage = bus.res_vm_pu
    # if self.unsatisfied_constraints or voltage < 0.95:
    #     return False
    # else:
    #     return True
    # pass



from abc import ABC, abstractmethod
from typing import Optional

from ..elements.base import GridElement


class Constraint(ABC):
    def __init__(self, element: GridElement):
        self._element: GridElement = element
        self.expected_value: Optional[float] = None
        self.violated_value: Optional[float] = None

        self._satisfied: bool = True

    def set_violation_values(self, expected_value, violated_value):
        self.violated_value = violated_value
        self.expected_value = expected_value

    @abstractmethod
    def check(self, time) -> bool:
        pass

    @abstractmethod
    def handle_violation(self):
        pass



"""This module contains a wrapper for pandapower grids."""

from typing import Dict, List, Union

import numpy as np
import pandapower as pp

from ..elements.bus import PPBus
from ..elements.line import PPLine
from ..elements.load import PPLoad
from ..elements.sgen import PPSgen
from ..elements.transformer import PPTransformer
from ..meta import ATTRIBUTE_MAP
from . import LOG
from .pp_grid import PPGrid
from .surrogate import SurrogateGrid


class PandapowerGrid:
    """A model for pandapower grids."""

    def __init__(self, params, grid_idx=0):
        self.entity_map = {}
        self.grid = None
        self.grid_idx = grid_idx
        self.has_profiles = False
        self.time_step = 0
        self.ids = {}
        self.cache = {}
        self.grid_type = None

        self.run_diagnostic = False
        self.lf_converged = False
        self._constraints_to_load = []
        self.constraints = {}
        self._lf_states: List[bool] = []
        self._include_slack_bus: bool = False

        self.setup(params)

    def setup(self, model_params):
        """Set up the grid model."""

        pp_params = model_params.setdefault("pp_params", {})

        if "surrogate_params" in pp_params:
            self.grid = SurrogateGrid(model_params["gridfile"], pp_params)
        else:
            self.grid = PPGrid(model_params["gridfile"], pp_params)

        self._include_slack_bus = model_params.get("include_slack_bus", False)
        self._load_grid_ids()
        self._load_entity_map()

        # To save some time during runtime
        self.run_powerflow(-1)

        self._constraints_to_load = model_params.get("constraints", [])
        self._load_constraints()

    def set_inputs(self, etype, idx, data):
        """Set input from other simulators."""
        etype = etype.lower()
        if etype not in ["load", "sgen", "trafo", "switch", "storage"]:
            LOG.info("Invalid etype %s. Skipping.", etype)
            return False

        for name, value in data.items():
            # Add try/except ?
            if value is None or np.isnan(value):
                # print(f"Skipping {etype}-{idx}-{name} because of None")
                continue
            # else:
            #     print(f"Setting {etype}-{idx}-{name}: {value}")

            if etype == "switch" and name == "closed":
                if not isinstance(value, bool):
                    if isinstance(value, float):
                        value = (
                            value >= 0.5
                        )  # workaround since no hybrid SAC available
                    else:
                        value = value != 0
            if etype == "trafo" and name == "tap_pos":
                if not isinstance(value, int):
                    if isinstance(value, float):
                        try:
                            minv = self.grid.get_value(etype, idx, "tap_min")
                            maxv = self.grid.get_value(etype, idx, "tap_max")
                        except KeyError:
                            value = 0
                            LOG.warning(
                                "Trying to access 'tap_min' and 'tap_max' but"
                                "those do not exist in the grid. Will set "
                                "value to 0."
                            )
                        else:
                            value = int(
                                round(
                                    minv
                                    + ((1.0 + value) / 2.0) * (maxv - minv)
                                )
                            )
                    else:
                        value = 0
            if name == "in_service":
                if not isinstance(value, bool):
                    if isinstance(value, float):
                        value = value >= 0.5
                    else:
                        value = value != 0

            self.grid.set_value(etype, idx, name, value)

            if etype in self.constraints:
                # Constraint can change the value
                setattr(self.constraints[etype][idx], name, value)

    def run_powerflow(self, time, max_iter=2):
        """Run the powerflow calculation."""
        if self.constraints:
            self._run_constraints(time)
        else:
            self._run_powerflow(time)

    def _run_powerflow(self, time):
        """Run the powerflow calculation."""
        try:
            self.grid.run_powerflow()
            self.lf_converged = True
        except pp.LoadflowNotConverged:
            LOG.info(
                "At step %d: Loadflow did not converge. Set "
                "*run_diagnostic* to True "
                "to run pandapower diagnostics.",
                time,
            )
            self.lf_converged = False

            if self.run_diagnostic:
                pp.diagnostic(self.grid)

        self.cache = {}

    def _run_constraints(self, time, max_iter=2):
        self._lf_states.append(self.lf_converged)  # previous state

        # Run once to check current state
        self._run_powerflow(time)

        if time < 0:
            # The first step is done before the simulation to make use
            # of numba's speed-up during the simulation
            return

        state_changed = False

        # Now constraints can change the input state if necessary
        # Constraints will definitively change the state of the grid
        # so another power flow calculation is required afterwards
        for key in ["trafo", "load", "sgen", "line", "bus"]:
            if key not in self.constraints:
                continue
            for element in self.constraints[key]:
                state_changed = element.step(time) or state_changed

        # Maybe more elements were put out of service than necessary
        # e.g., when the critical element is the last one checked.
        # Performing another iteration allows elements to switch on
        # again if no constraints are violated.
        if max_iter > 1:
            return self._run_constraints(time, max_iter - 1)

        self._run_powerflow(time)
        if state_changed:
            if not self._lf_states[0] and self.lf_converged:
                LOG.info(f"At step {time}: Constraints fixed failing LF.")
            if not self._lf_states[0] and not self.lf_converged:
                LOG.info(f"At step {time}: LF still not converging.")
            if self._lf_states[0] and not self.lf_converged:
                LOG.info(f"At step {time}: LF broke due to constraints.")
            if self._lf_states[0] and self.lf_converged:
                LOG.info(f"At step {time}: LF still converging.")

        self._lf_states = []

    def get_outputs(self) -> Dict[str, Dict[str, Union[float, int, bool]]]:
        if self.cache:
            return self.cache

        for eid, attrs in self.entity_map.items():
            data = {}
            for otype, outputs in ATTRIBUTE_MAP[attrs["etype"]].items():
                try:
                    element = self.grid.get_value(otype, attrs["idx"])
                except IndexError:
                    LOG.exception(
                        f"Failed to get element {otype} from index "
                        f"{attrs['idx']}"
                    )
                for output in outputs:
                    data[output[0]] = (
                        _convert_bool(element[output[0]])
                        if self.lf_converged
                        else 0
                    )
            self.cache[eid] = data

        return self.cache

    def to_json(self):
        return pp.to_json(self.grid)

    def _load_grid_ids(self):
        """Create a dictionary containing the names of the components.

        Use generic names and map to actual names?

        """
        self.ids["slack"] = self.grid.get_value(
            "ext_grid", attr="bus"
        ).to_dict()
        self.ids["bus"] = self.grid.get_value("bus", attr="name").to_dict()
        self.ids["load"] = self.grid.get_value("load", attr="name").to_dict()
        self.ids["sgen"] = self.grid.get_value("sgen", attr="name").to_dict()
        self.ids["line"] = self.grid.get_value("line", attr="name").to_dict()
        self.ids["trafo"] = self.grid.get_value("trafo", attr="name").to_dict()
        self.ids["switch"] = self.grid.get_value(
            "switch", attr="name"
        ).to_dict()
        self.ids["storage"] = self.grid.get_value(
            "storage", attr="name"
        ).to_dict()

    def _load_entity_map(self):
        """Load the entity map for the mosaik simulator."""

        self._get_slack()
        self._get_buses()
        self._get_loads()
        self._get_sgens()
        self._get_lines()
        self._get_trafos()
        self._get_switches()
        self._get_storages()

    def _get_slack(self):
        """Create an entity for the slack bus."""
        for idx in self.ids["slack"]:
            element = self.grid.get_value("ext_grid", idx)
            eid = self._create_eid(
                "ext_grid", idx, self.grid.get_value("ext_grid", idx, "bus")
            )

            self.entity_map[eid] = {
                "etype": "Ext_grid",
                "idx": int(idx),
                "static": {
                    "name": element["name"],
                    "vm_pu": float(element["vm_pu"]),
                    "va_degree": float(element["va_degree"]),
                },
            }

    def _is_slack_bus(self, bus_id):
        for bus in self.ids["slack"].values():
            if bus == bus_id:
                return True

        return False

    def _get_buses(self):
        """Create entities for buses."""
        for idx in self.ids["bus"]:
            if not self._include_slack_bus and self._is_slack_bus(idx):
                continue

            element = self.grid.get_value("bus", idx)
            eid = self._create_eid("bus", idx)
            self.entity_map[eid] = {
                "etype": "Bus",
                "idx": int(idx),
                "static": {
                    "name": element["name"],
                    "vn_kv": float(element["vn_kv"]),
                },
            }

    def _get_loads(self):
        """Create entities for loads."""
        for idx in self.ids["load"]:
            element = self.grid.get_value("load", idx)
            eid = self._create_eid("load", idx, element["bus"])
            bid = self._create_eid("bus", element["bus"])
            element_data = element.to_dict()

            keys_to_del = [
                "profile",
                "voltLvl",
                "const_z_percent",
                "const_i_percent",
                "min_q_mvar",
                "min_p_mw",
                "max_q_mvar",
                "max_p_mw",
            ]
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Load",
                "idx": int(idx),
                "static": element_data_static,
                "related": [bid],
            }

    def _get_sgens(self):
        """Create entities for sgens."""
        for idx in self.ids["sgen"]:
            element = self.grid.get_value("sgen", idx)
            eid = self._create_eid("sgen", idx, element["bus"])
            bid = self._create_eid("bus", element["bus"])
            element_data = element.to_dict()

            keys_to_del = [
                "profile",
                "voltLvl",
                "min_q_mvar",
                "min_p_mw",
                "max_q_mvar",
                "max_p_mw",
            ]
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Sgen",
                "idx": int(idx),
                "static": element_data_static,
                "related": [bid],
            }

    def _get_lines(self):
        """Create entities for lines."""
        for idx in self.ids["line"]:
            element = self.grid.get_value("line", idx)
            eid = self._create_eid("line", idx)
            fbid = self._create_eid("bus", element["from_bus"])
            tbid = self._create_eid("bus", element["to_bus"])

            element_data = element.to_dict()
            keys_to_del = ["from_bus", "to_bus"]
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Line",
                "idx": int(idx),
                "static": element_data_static,
                "related": [fbid, tbid],
            }

    def _get_trafos(self):
        """Create entities for trafos."""
        for idx in self.ids["trafo"]:
            element = self.grid.get_value("trafo", idx)
            eid = self._create_eid("trafo", idx)
            hv_bid = self._create_eid("bus", element["hv_bus"])
            lv_bid = self._create_eid("bus", element["lv_bus"])

            element_data = element.to_dict()
            keys_to_del = ["hv_bus", "lv_bus"]
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Trafo",
                "idx": int(idx),
                "static": element_data_static,
                "related": [hv_bid, lv_bid],
            }

    def _get_switches(self):
        """Create entities for switches."""
        for idx in self.ids["switch"]:
            element = self.grid.get_value("switch", idx)
            eid = self._create_eid("switch", idx)
            bid = self._create_eid("bus", element["bus"])

            if element["et"] == "l":
                oid = self._create_eid("line", element["element"])
            elif element["et"] == "t":
                oid = self._create_eid("trafo", element["element"])
            elif element["et"] == "b":
                oid = self._create_eid("bus", element["element"])

            element_data = element.to_dict()
            keys_to_del = ["element"]
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Switch",
                "idx": int(idx),
                "static": element_data_static,
                "related": [bid, oid],
            }

    def _get_storages(self):
        """Create entities for storages."""
        for idx in self.ids.get("storage", list()):
            element = self.grid.get_value("storage", idx)
            eid = self._create_eid("storage", idx, element["bus"])
            bid = self._create_eid("bus", element["bus"])
            element_data = element.to_dict()

            keys_to_del = []
            element_data_static = {
                key: element_data[key]
                for key in element_data
                if key not in keys_to_del
            }

            self.entity_map[eid] = {
                "etype": "Storage",
                "idx": int(idx),
                "static": element_data_static,
                "related": [bid],
            }

    def _create_eid(self, name, idx, bus_id=None):
        eid = f"{self.grid_idx}-{name}-{idx}"
        if bus_id is not None:
            eid = f"{eid}-{bus_id}"
        return eid

    def finalize(self):
        self.grid.finalize()

    def _load_constraints(self):
        for constr in self._constraints_to_load:
            etype, value = constr
            self.constraints.setdefault(etype, list())
            for idx in range(len(self.grid.get_value(etype))):
                self.constraints[etype].append(self._create(etype, idx, value))

    def _create(self, etype, index, value):
        if etype == "trafo":
            clazz = PPTransformer
        # if classname == PPTransformer:
        #     etype = self.grid[classname.pp_key()]["std_type"][index]

        #     return PPTransformer(index, self.grid)
        # TODO: elif other elements
        elif etype == "bus":
            clazz = PPBus
        elif etype == "load":
            clazz = PPLoad
        elif etype == "sgen":
            clazz = PPSgen
        elif etype == "line":
            clazz = PPLine
        return clazz(index, self.grid, value)


def _convert_bool(val):
    if isinstance(val, bool):
        val = 1 if val else 0
    try:
        if not isinstance(val, str):
            if np.isnan(val):
                val = 0
    except TypeError:
        print(f"value: {val} ({type(val)})")
        raise
    return val



from .voltage_change import ConstraintVoltageChange


class ConstraintSgenVoltageChange(ConstraintVoltageChange):
    """Voltage change of a sgen is not allowed to rise over 5%."""

    def __init__(self, element, expected_value):
        super().__init__(element)

        self._expected_value = expected_value

    def handle_violation(self):
        # Undervoltages should be handled by loads
        if self._time_voltages[-1].value > 1.0:
            self._element.in_service = False
        self._element.current_bus_voltage = self._time_voltages[-1].value



from importlib import import_module

import pandapower as pp
import pandapower.networks as pn
import simbench as sb

from ..custom import bhv, midaslv, midasmv


class PPGrid:
    def __init__(self, grid_name, grid_params):
        self._grid = None
        self.plot_type: str = "cigre"
        self._numba = not grid_params.get("disable_numba", False)

        if grid_name == "midasmv":
            self._grid = midasmv.build_grid(**grid_params)
        elif grid_name == "midaslv":
            self._grid = midaslv.build_grid(**grid_params)
        elif grid_name == "bhv":
            self._grid = bhv.build_grid(**grid_params)
            self.plot_type = "other"
        elif grid_name == "cigre_hv":
            self._grid = pn.create_cigre_network_hv(**grid_params)
        elif grid_name == "cigre_mv":
            self._grid = pn.create_cigre_network_mv(**grid_params)
        elif grid_name == "cigre_lv":
            self._grid = pn.create_cigre_network_lv(**grid_params)
        elif grid_name in ("oberrhein", "mv_obverrhein"):
            self._grid = pn.mv_oberrhein()
        elif grid_name.endswith(".json"):
            self._grid = pp.from_json(grid_name)
            self.plot_type = "other"
        elif grid_name.endswith(".xlsx"):
            self._grid = pp.from_excel(grid_name)
        elif self._load_simbench(grid_name):
            self.plot_type = "simbench"
        elif "." in grid_name:
            if ":" in grid_name:
                mod, clazz = grid_name.split(":")
            else:
                mod, clazz = grid_name.rsplit(".", 1)
            mod = import_module(mod)
            self._grid = getattr(mod, clazz)()
            self.plot_type = "other"
        else:
            self._grid = getattr(pn, grid_name)()
            self.plot_type = "other"

    def set_value(self, etype, idx, attr, val):
        self._grid[etype].at[idx, attr] = val

    def run_powerflow(self):
        pp.runpp(self._grid, numba=self._numba)

    def get_value(self, etype, idx=None, attr=None):
        if idx is None and attr is None:
            return self._grid[etype]
        elif attr is None:
            return self._grid[etype].iloc[idx]
        elif idx is None:
            return self._grid[etype][attr]
        else:
            return self._grid[etype].at[idx, attr]

    def get_outputs(self):
        raise NotImplementedError

    def to_json(self):
        return pp.to_json(self._grid)

    def _load_simbench(self, gridfile):
        """Try to load a simbench grid.

        Importing the simbench module is done here because that takes
        a few seconds to load, which are wasted if simbench is not used
        at all.

        """

        try:
            self._grid = sb.get_simbench_net(gridfile)
        except ValueError:
            return False

        return True

    def finalize(self):
        pass



"""This module contains the MIDAS powergrid upgrade."""

import logging
import os
from importlib import import_module

from midas.scenario.upgrade_module import UpgradeModule
from midas.util.dict_util import (
    set_default_bool,
    set_default_float,
    set_default_int,
)

from .analysis.grid import analyze_grid
from .meta import ATTRIBUTE_MAP

LOG = logging.getLogger(__name__)


class PowergridModule(UpgradeModule):
    """The MIDAS powergrid update module."""

    def __init__(self):
        super().__init__(
            module_name="powergrid",
            default_scope_name="midasmv",
            default_sim_config_name="Powergrid",
            default_import_str=(
                "midas_powergrid.simulator:PandapowerSimulator"
            ),
            default_cmd_str=(
                "%(python)s -m midas_powergrid.simulator %(addr)s"
            ),
            log=LOG,
        )

        self.models = {}
        for entity, data in ATTRIBUTE_MAP.items():
            self.models.setdefault(entity, [])
            for attrs in data.values():
                for attr in attrs:
                    self.models[entity].append(attr)

        self._double_actuator_values: bool = False

    def check_module_params(self, module_params: dict):
        """Check the module params for this upgrade."""

        set_default_bool(module_params, "plotting", False)
        module_params.setdefault(
            "plot_path",
            os.path.abspath(
                os.path.join(self.scenario.base.output_path, "plots")
            ),
        )
        set_default_bool(module_params, "save_grid_json", False)
        module_params.setdefault("constraints", [])
        set_default_float(module_params, "actuator_multiplier", 1.0)
        set_default_bool(module_params, "include_slack_bus", False)

    def check_sim_params(self, module_params: dict):
        """Check simulator params for a certain simulator."""

        self.sim_params.setdefault("gridfile", self.default_scope_name)
        self.sim_params.setdefault("grid_params", {})
        set_default_bool(
            self.sim_params, "plotting", module_params["plotting"]
        )
        self.sim_params.setdefault("plot_path", module_params["plot_path"])
        set_default_bool(
            self.sim_params, "save_grid_json", module_params["save_grid_json"]
        )
        self.sim_params.setdefault("constraints", module_params["constraints"])
        set_default_float(
            self.sim_params,
            "actuator_multiplier",
            module_params["actuator_multiplier"],
        )
        self.sim_params.setdefault("grid_mapping", {})
        set_default_bool(
            self.sim_params,
            "include_slack_bus",
            module_params["include_slack_bus"],
        )

    def start_models(self):
        """Start all models for this simulator.

        Since we want the grids to be able to be interconnected with
        each other, each grid model should have its own simulator.

        Parameters
        ----------
        sim_name : str
            The sim name, not to be confused with *sim_name* for
            mosaik's *sim_config*. **This** sim_name is the simulator
            key in the configuration yaml file.

        """
        model_key = self.scenario.generate_model_key(self)

        model_params = {
            "gridfile": self.sim_params["gridfile"],
            "pp_params": self.sim_params["grid_params"],
            "constraints": self.sim_params["constraints"],
            "include_slack_bus": self.sim_params["include_slack_bus"],
        }

        self.start_model(model_key, "Grid", model_params)

        if ":" in self.default_import_str:
            mod, clazz = self.default_import_str.split(":")
        else:
            mod, clazz = self.default_import_str.rsplit(".", 1)
        mod = import_module(mod)

        sim_dummy = getattr(mod, clazz)()
        sim_dummy.init(self.sid, **self.sim_params)
        entities = sim_dummy.create(1, "Grid", **model_params)
        grid_obj = sim_dummy.models[entities[0]["eid"]].grid
        self.scenario.add_to_model_extra_data(
            model_key, self.sim_key, {"grid": grid_obj}
        )
        self.scenario.add_to_world_state("grid_json", grid_obj.to_json())
        # Backwards compatibility
        # Only works for midasmv anyways
        self._double_actuator_values = model_params["pp_params"].get(
            "double_actuator_values", False
        )
        if self._double_actuator_values:
            self.sim_params["actuator_multiplier"] = 2.0
        #

        # for entity in self.scenario.get_model(
        #     model_key, self.sim_key
        # ).children:
        for entity in entities[0]["children"]:
            eid = entity["eid"]
            parts = eid.split("-")
            child_key = f"{model_key}"
            for part in parts[1:]:
                child_key += f"_{part}"

            self.scenario.script.model_start.append(
                f"{child_key} = [e for e in {model_key}.children "
                f'if e.eid == "{eid}"][0]\n'
            )
            self.scenario.add_model(
                child_key,
                self.sim_key,
                f"{entity['type']}",
                {},
                f"{self.sid}.{eid}",
                f"{self.sid}.{entities[0]['eid']}",
            )

    def connect(self, *args):
        # Nothing to do so far
        # Maybe to other grids in the future?
        pass

    def connect_to_db(self):
        """Add connections to the database."""

        grid_key = self.scenario.generate_model_key(self)
        db_key = self.scenario.find_first_model("store", "database")[0]

        for key, entity in self.scenario.get_models(self.sim_key).items():
            if grid_key not in key:
                continue
            mod_key = key

            # We only want to connect those attributes that are present
            # in the grid. That's why we iterate over existing entities
            # and not simply use the models defined above.
            if entity["name"] in self.models:
                self.connect_entities(
                    mod_key,
                    db_key,
                    [a[0] for a in self.models[entity["name"]]],
                )

        additional_attrs = ["health"]
        if self.sim_params["save_grid_json"]:
            additional_attrs.append("grid_json")

        self.connect_entities(grid_key, db_key, additional_attrs)

    def get_sensors(self):
        models = self.scenario.get_models(self.sim_key)
        for model_key, config in models.items():
            if config["name"] in self.models:
                for attr in self.models[config["name"]]:
                    name, dtype, low, high = attr

                    if dtype == "bool":
                        space = f"Discrete({high})"
                    elif dtype == "int":
                        space = _int_box(low, high)
                    else:
                        space = _float_box(low, high)

                    self.scenario.sensors.append(
                        {"uid": f"{config['full_id']}.{name}", "space": space}
                    )

        grid_key = self.scenario.generate_model_key(self)
        grid = self.scenario.get_model(grid_key)
        self.scenario.sensors.append(
            {"uid": f"{grid['full_id']}.health", "space": _float_box(0, 1.2)}
        )
        # self.scenario.sensors.append(
        #     {
        #         "uid": f"{grid.full_id}.grid_json",
        #         "space": ("Box(low=0, high=1, shape=(), dtype=np.float32)"),
        #     }
        # )

    def get_actuators(self):
        grid_key = self.scenario.generate_model_key(self)
        grid = self.scenario.get_model(grid_key)
        models = self.scenario.get_models(self.sim_key)
        for model_key, config in models.items():
            if config["name"] == "Trafo":
                self.scenario.actuators.append(
                    {
                        "uid": f"{config['full_id']}.tap_pos",
                        "space": _int_box(
                            -10, 10
                        ),  # FIXME get trafo values from grid
                    }
                )
            if config["name"] == "Switch":
                self.scenario.actuators.append(
                    {"uid": f"{config['full_id']}.closed", "space": _int_box()}
                )
            if config["name"] in ("Load", "Sgen"):
                q_min = 0
                p_min = 0
                try:
                    pp_grid = grid["extra_data"]["grid"]
                    # try:
                    #     pp_grid = list(
                    #         self.scenario.get_sim(
                    #             self.sim_key
                    #         )._sim._inst.models.values()
                    #     )[0].grid
                    # except AttributeError:
                    #     pp_grid = list(
                    #         self.scenario.get_sim(
                    #             self.sim_key
                    #         )._proxy.sim.models.values()
                    #     )[0].grid

                    _, etype, eidx, _ = config["eid"].split("-")
                    try:
                        p_max = pp_grid.get_value(etype, int(eidx), "max_p_mw")
                        p_min = pp_grid.get_value(etype, int(eidx), "min_p_mw")
                    except KeyError:
                        LOG.debug(
                            "No max_p_mw or min_p_mw found, using p_mw as "
                            "p_max and set p_min to 0"
                        )
                        p_max = pp_grid.get_value(etype, int(eidx), "p_mw")

                    try:
                        q_max = pp_grid.get_value(
                            etype, int(eidx), "max_q_mvar"
                        )
                        q_min = pp_grid.get_value(
                            etype, int(eidx), "min_q_mvar"
                        )
                    except KeyError:
                        LOG.debug(
                            "No max_p_mw or min_p_mw found, using q_mvar as "
                            "q_max and set q_min to 0"
                        )
                        q_max = pp_grid.get_value(etype, int(eidx), "q_mvar")

                except Exception as err:
                    # print(err)
                    LOG.warning(
                        "Caught an error composing actuator information for "
                        f"{etype} with eid {config['eid']}: {type(err), err}. "
                        "Continuing with p_max=0.5, q_max=0.5, and "
                        "p_min=q_min=0.",
                        exc_info=err,
                    )
                    p_max = 0.5
                    q_max = 0.5

                p_max *= self.sim_params["actuator_multiplier"]
                q_max *= self.sim_params["actuator_multiplier"]

                if q_max < q_min:
                    q_min = q_max
                    q_max = 0

                if p_min < p_max:
                    self.scenario.actuators.append(
                        {
                            "uid": f"{config['full_id']}.p_mw",
                            "space": _float_box(p_min, p_max),
                        }
                    )
                if q_min < q_max:
                    self.scenario.actuators.append(
                        {
                            "uid": f"{config['full_id']}.q_mvar",
                            "space": _float_box(q_min, q_max),
                        }
                    )

    def analyze(self, name, data, output_folder, start, end, step_size, full):
        grid_sim_keys = [
            sim_key for sim_key in data.keys() if "Powergrid" in sim_key
        ]
        for sim_key in grid_sim_keys:
            grid_data = data[sim_key]
            if start > 0:
                grid_data = grid_data.iloc[start:]
            if end > 0:
                grid_data = grid_data.iloc[:end]

            analyze_grid(
                grid_data,
                step_size,
                f"{name}-{sim_key.replace('/', '')}",
                output_folder,
                full,
            )

    def download(self, *args):
        # No downloads, suppress logging output
        pass


def _float_box(low=0.0, high=1.0):
    return f"Box(low={low}, high={high}, shape=(), dtype=np.float32)"


def _int_box(low=0, high=1):
    return f"Box(low={low}, high={high}, shape=(), dtype=np.int32)"



import collections
from datetime import timedelta

from .base import Constraint

TimedVoltage = collections.namedtuple("TimedVoltage", ["time", "value"])


class ConstraintVoltageChange(Constraint):
    def __init__(self, element):
        super().__init__(element)

        self._expected_value = 0.02
        self._time_frame = timedelta(minutes=1)
        self._time_voltages = list()
        self._bus = self._element.grid.get_value(
            self._element.pp_key(), self._element.index, "bus"
        )

    def check(self, time) -> bool:
        self._satisfied = True
        voltage = self._element.grid.get_value("res_bus", self._bus, "vm_pu")
        self._time_voltages = [
            entry
            for entry in self._time_voltages
            if time - entry.time < self._time_frame.seconds
        ]
        current_t_v = TimedVoltage(time=time, value=voltage)
        self._time_voltages.append(current_t_v)

        voltage_values = [entry.value for entry in self._time_voltages]
        min_voltage = min(voltage_values)
        max_voltage = max(voltage_values)

        voltage_change_percent = abs(max_voltage - min_voltage) / min_voltage
        voltage_change_percent = round(voltage_change_percent, 6)

        if voltage_change_percent > self._expected_value:
            self._satisfied = False
            self._violated_value = voltage_change_percent

        return self._satisfied

    def handle_violation(self):
        self._element.in_service = False



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from midas.util import report_util

from .bus import analyze_buses
from .power import analyze_power


def analyze_grid(data, step_size, name, output_path, full_report):
    plot_path = os.path.join(output_path, name.rsplit("-", 1)[1]).replace(
        "__", "_"
    )
    os.makedirs(plot_path, exist_ok=True)

    ef = step_size / 3_600
    report_content = list()
    bus_data = data[[col for col in data.columns if "bus" in col]]
    score = analyze_buses(
        bus_data, report_content, name, plot_path, full_report
    )

    def is_ext_grid(col):
        return "ext_grid" in col or "slack" in col

    ext_grid_data = data[[col for col in data.columns if is_ext_grid(col)]]
    extgrid_totals = analyze_power(
        ext_grid_data,
        step_size,
        report_content,
        plot_path,
        full_report,
        {"name": name, "topic": "ExtGrid", "total_name": "0-extgrids"},
    )
    load_data = data[[col for col in data.columns if "load__" in col]]
    load_totals = analyze_power(
        load_data,
        step_size,
        report_content,
        plot_path,
        full_report,
        {"name": name, "topic": "Load", "total_name": "0-loads"},
    )

    sgen_data = data[[col for col in data.columns if "sgen" in col]]
    sgen_totals = analyze_power(
        sgen_data,
        step_size,
        report_content,
        plot_path,
        full_report,
        {"name": name, "topic": "Sgen", "total_name": "0-sgens"},
    )

    storage_data = data[[col for col in data.columns if "storage" in col]]
    if not storage_data.empty:
        storage_totals = analyze_power(
            storage_data,
            step_size,
            report_content,
            plot_path,
            full_report,
            {"name": name, "topic": "Storage", "total_name": "0-storages"},
        )
        # Storages work in consumer system,
        # i.e., charging > 0, discharging < 0
        load_sgen_diff = pd.DataFrame(
            {
                "load+storage-sgen.p_mw": load_totals[0]
                + storage_totals[0]
                - sgen_totals[0],
                "load+storage-sgen.q_mvar": load_totals[1]
                + storage_totals[1]
                - sgen_totals[1],
            }
        )
    else:
        load_sgen_diff = pd.DataFrame(
            {
                "load-sgen.p_mw": load_totals[0] - sgen_totals[0],
                "load-sgen.q_mvar": load_totals[1] - sgen_totals[1],
            }
        )

    analyze_power(
        load_sgen_diff,
        step_size,
        report_content,
        plot_path,
        False,
        {"name": name, "topic": "Load-Balance", "total_name": "0-load-sgen"},
    )

    report_path = os.path.join(
        output_path, f"{name.replace('__', '_')}_report.md"
    )
    report_file = open(report_path, "w")

    if load_totals[0].sum() == 0:
        energy_suffiency = 0
    else:
        energy_suffiency = 100 * sgen_totals[0].sum() / load_totals[0].sum()

    if sgen_totals[3] == 0:
        full_load_hours = 0
    else:
        full_load_hours = sgen_totals[0].sum() * ef / sgen_totals[3]

    report_file.write(
        f"# Analysis of {name}\n\n## Summary\n\n"
        f"* bus health: {score:.2f} %\n"
        f"* active energy sufficiency: {energy_suffiency:.2f} %\n"
        # "* reactive energy sufficiency: "
        # f"{100*sgen_totals[1].sum()/load_totals[1].sum():.2f} %\n"
        f"\n## Demand and Supply\n\n"
        f"* total active energy demand: {load_totals[0].sum() * ef:.2f} MWh\n"
        f"* total active energy supply: {sgen_totals[0].sum() * ef:.2f} MWh "
        f"or about {full_load_hours:.2f} full load hours\n"
        "* extg. active energy supply: "
        f"{extgrid_totals[0].sum()*ef:.2f} MWh\n"
        "* total reactive energy demand: "
        f"{load_totals[1].sum()*ef:.2f} MVArh\n"
        "* total reactive energy supply: "
        f"{sgen_totals[1].sum()*ef:.2f} MVArh\n"
        f"* extg. reactive energy supply: {extgrid_totals[1].sum()*ef:.2f} "
        "MVArh\n"
        f"* total apparent energy demand: {load_totals[2]:.2f} MVAh\n"
        f"* total apparent energy supply: {sgen_totals[2]:.2f} MVAh\n"
        f"* extg. apparent energy supply: {extgrid_totals[2]:.2f} MVAh\n\n"
    )

    for line in report_content:
        report_file.write(f"{line}\n")
    report_file.close()

    report_util.convert_markdown(report_path)


def analyze_line(data, report_file, name, output_path):
    load_percent = np.array([data[key]["loading_percent"] for key in data])

    data["line_avg"] = {"loading_percent": load_percent.mean(axis=0)}

    for key, vals in data.items():
        load_percent = np.array(vals["loading_percent"])

        annual = np.sort(load_percent)[::-1]
        too_high10 = (annual > 120).sum()
        too_high4 = (annual > 60).sum()

        if too_high10 > 0:
            report_file.write(f"[{key}] {too_high10} values > 120\n")
        if too_high4 > 0:
            report_file.write(f"[{key}] {too_high4} values > 60\n")

        fig, ax = plt.subplots(1, 1, figsize=(12, 8))
        ax.plot(annual)
        ax.axhline(y=120, color="red")
        ax.axhline(y=60, linestyle="--", color="red")
        ax.set_title(f"{key}")
        ax.set_ylabel("Line load percentage")
        ax.set_xlabel("time (15 minute steps)")
        plt.savefig(
            os.path.join(
                output_path, f"{name}_{key}_load_percentage_annual.png"
            ),
            dpi=300,
            bbox_inches="tight",
        )
        plt.close()
